//! \file filesdb.cc Реализация класса работы с БД
#include <fstream>
#include <sstream>
#include <algorithm>
#include <errno.h>
#include <time.h>
#include <stdlib.h>
#include <pthread.h>
#include "filesdb.h"
#include <iostream>

//! Оператор записи объекта sync_object в поток ostream
/**
 * \param s ссылка на объект потока
 * \param f объект sync_object
 */
ostream& operator<<(ostream& s,sync_object f)
{
	string data = f.getStringData();
	unsigned sz=data.size();
	s.write((char*)&sz,sizeof(sz));
	s.write(data.c_str(),data.size());
	return s;
}

//! Оператор чтения объекта sync_object из потока istream
/**
 * \param s ссылка на объект потока
 * \param f ссылка на объект sync_object
 */
istream& operator>>(istream& s,sync_object &f)
{
	unsigned sz;
	char* buf;

	s.exceptions(ios_base::badbit|ios_base::eofbit|ios_base::failbit);
	s.read((char*)&sz,sizeof(sz));			
	buf=new char[sz];
	s.read(buf,sz);
	f=sync_object(string(buf,sz));
	delete []buf;

	return s;
}


//! Конструктор класса 
/**
 * \param fpath путь к БД
 */
filedb::filedb(string fpath):path(fpath)
{
	pthread_mutex_init(&m_lock,0);
}

//! Конструктор класса по умолчанию
filedb::filedb()
{
	pthread_mutex_init(&m_lock,0);
}

//! Деструктор класса
filedb::~filedb()
{
	pthread_mutex_destroy(&m_lock);
	for(vector<sync_object*>::iterator i=m_db.begin(); i!=m_db.end(); i++)
		delete *i;
}

//! Установка пути к БД
/**
 * \param fpath путь к БД
 */
void filedb::dbset(string fpath)
{
	path=fpath;
}
//! Блокировка БД
int filedb::lock()
{
	return pthread_mutex_lock(&m_lock);
}
//! Разблокировка БД
int filedb::unlock()
{
	return pthread_mutex_unlock(&m_lock);
}
//! Загрузка из файла БД в вектор класса filedb 
/**
 * \return 0 - успех, иначе - неудача
 */
int filedb::load()
{
	sync_object buf;
	ifstream f(path.c_str());
	if(!f)
		return -EACCES;
	m_db.clear();
	try{
		while(f>>buf){
			buf.setDB(this);
			m_db.push_back(new sync_object(buf));
		}
	}
	catch(ios_base::failure){
	}
	f.close();
	return 0;
}

//! Сохранение в файл БД из внутренненго вектор класса filedb 
/**
 * \return 0 - успех, иначе - неудача
 */
int filedb::save()
{
	ofstream f(path.c_str(),ios_base::out|ios_base::trunc);
	if(!f)
		return -EACCES;
	for(vector<sync_object*>::iterator i=m_db.begin(); i!=m_db.end(); i++)
		f<<**i;
	f.close();
	return 0;
}

//! Добавление объекта синхронизации с БД
/**
 * \param obj объект синхронизации
 * \return указатель на добавленный объект
 */
sync_object* filedb::insert(sync_object obj)
{
	for(vector<sync_object*>::iterator i=m_db.begin(); i!=m_db.end(); i++)
        {
		if(**i==obj) {
			return *i; }
		if((*i)->getUName() == obj.getUName())
			return *i;
	}
	obj.setDB(this);
	sync_object* f= new sync_object(obj);
	m_db.push_back(f);
	return f;
}

//! Добавление файла в БД
/**
 * \param fpath путь к файлу
 * \return указатель на добавленный объект
 */
sync_object* filedb::insert(string fpath)
{
	for(vector<sync_object*>::iterator i=m_db.begin(); i!=m_db.end(); i++)
	{
		if(**i==fpath)
		{
			if((*i)->toDelete())
			{
				lock();
				(*i)->undelete();
				save();
				unlock();
				return *i;
			}
			return *i;
		};
	};
	lock();
	sync_object* f= new sync_object(fpath,string());
	f->setDB(this);
	m_db.push_back(f);
	save();
	unlock();
	return f;
}
//! Удаление удаленных с сервера объектов из БД
void filedb::delDeleted()
{
	for(vector<sync_object*>::iterator i=m_db.begin(); i!=m_db.end(); ) {
		if((*i)->canDelete()){
			delete *i;
			i=m_db.erase(i);
		}else 
			i++;
	}
}
//! Пометка объекта синхронизации к удалению
/**
 * \param fpath путь к объекту
 * \return 0 - успех, иначе - неудача
 */
int filedb::remove(string fpath)
{
	vector<sync_object*>::iterator i;
	lock();
	for(i=m_db.begin(); i!=m_db.end(); i++)
		if(**i==fpath)
			break;
	if(i==m_db.end()){
		unlock();
		return -1;
	}
	(*i)->setDelete();
	save();
	unlock();
	return 0;
}
//! Удаление объекта из БД
/**
 * \param obj указатель на объект синхронизации
 */
int filedb::erase(sync_object* obj)
{
	vector<sync_object*>::iterator i= std::find(m_db.begin(),m_db.end(),obj);
	delete *i;
	m_db.erase(i);
	return 0;
}
//! Проход по всем объектам синхронизации из БД
/**
 * Получает указатель элемента следующего за iter и возвращает его.
 * Если iter=NULL,то возвращает указатель на первый элемент.
 * \param iter ссылка на указатель объекта синхронизации
 * \return 0 - конец списка, 1 - указатель получен успешно
 */
int filedb::nextall(sync_object*& iter)
{
	vector<sync_object*>::iterator i= std::find(m_db.begin(),m_db.end(),iter);
	if(i!=m_db.end())
		++i;
	if(iter==NULL)
		i=m_db.begin();
	if(i==m_db.end())
		return 0;
	iter=*i;
	return 1;
}

//! Проход по всем объектам синхронизации первого уровня из БД 
/**
 * Получает указатель элемента следующего за iter и возвращает его.
 * Если iter=NULL,то возвращает указатель на первый элемент.
 * \param iter ссылка на указатель объекта синхронизации
 * \return 0 - конец списка, 1 - указатель получен успешно
 */
int filedb::next(sync_object*& iter)
{
	vector<sync_object*>::iterator i= std::find(m_db.begin(),m_db.end(),iter);
	while(i!=m_db.end()){
		++i;
		if(i!=m_db.end()){
			if(!(*i)->level())
				break;
		}
	}
	if(iter==NULL)
		i=m_db.begin();
	if(i==m_db.end())
		return 0;
	iter=*i;
	return 1;
}
//! Добавление в БД объектов из вектора srv_files
/**
 * Добавляет в БД все объекты из вектора, кроме тех, которые уже есть 
 * в БД.
 * \param srv_files вектор объектов для добавления в БД
 */
int filedb::combine(vector<sync_object> srv_files)
{
	for(vector<sync_object>::iterator i=srv_files.begin();i!= srv_files.end(); i++){
		if(i->valid())
			insert(*i);
	}
	save();
	return 0;
}
//! Поиск объекта в классе БД по его пути
/**
 * \param fpath путь к файлу объекта
 * \return указатель на объект синхронизации, если объект найден, иначе NULL
 */
sync_object* filedb::find(string fpath)
{
	for(vector<sync_object*>::iterator i=m_db.begin(); i!=m_db.end(); i++)
		if(**i==fpath)
			return *i;
	return NULL;
}

//! Поиск объекта в классе БД по watch дескриптору inotify
/**
 * \param wd watch дескриптор inotify
 * \return указатель на объект синхронизации, если объект найден, иначе NULL
 */
sync_object* filedb::find_wd(int wd)
{
	for(vector<sync_object*>::iterator i=m_db.begin(); i!=m_db.end(); i++)
		if((*i)->get_ifd() == wd)
			return *i;
	return NULL;
}


//! Поиск объектов в классе БД по родительскому каталогу
/**
 * \param fpath путь к родительскому каталогу
 * \return вектор указателей на объекты синхронизации, находящиеся в данном каталоге
 */
vector<sync_object*> filedb::findDir(string fpath)
{
	vector<sync_object*> ret;
	for(vector<sync_object*>::iterator i=m_db.begin(); i!=m_db.end(); i++) {
		if((*i)->inDir(fpath)) {
			ret.push_back(*i);
		}
	}
	return ret;
}


